/*
  $Id: DataCont.h 4269 2009-02-09 20:08:05Z rares $

  Copyright (C) 2009 by The Regents of the University of California

  Redistribution of this file is permitted under the terms of the
  BSD license

  Date: 09/01/2005
  Author: Rares Vernica <rares (at) ics.uci.edu> 
*/

#ifndef __DataCont_h
#define __DataCont_h

#include <iostream>
#include <queue>
#include <set>
#include <vector>

using namespace std;

#include "DataStruct.h"
#include "DataSetUtil.h"

// --- compare

struct lessPair
{
  inline bool operator()(const PairRegion& x, const PairRegion& y)
  {
    return dist(x)<dist(y);
  }
  inline bool operator()(const PairRegionID& x, const PairRegionID& y)
  {
    return dist(x)<dist(y);
  }
};

struct greaterPair
{
  inline bool operator()(const PairRegion& x, const PairRegion& y)
  {
    return dist(x)>dist(y);
  }
  inline bool operator()(const PairRegionID& x, const PairRegionID& y)
  {
    return dist(x)>dist(y);
  }
};

struct lessPairScore
{
  inline bool operator()(const PairPoint& x, const PairPoint& y)
  {
    return scoreMin(x)<scoreMin(y);
  }
  inline bool operator()(const PairRegionID& x, const PairRegionID& y)
  {
    return scoreMin(x)<scoreMin(y);
  }
};

struct greaterPairScore
{
  inline bool operator()(const PairRegionID& x, const PairRegionID& y)
  {
    return scoreMin(x)>scoreMin(y);
  }
};

/* ----- vect ----- */

typedef vector<Point> VectPoint;
typedef VectPoint::iterator VectPointIt;

typedef vector<PairPoint> VectPairPoint;
typedef VectPairPoint::iterator VectPairPointIt;

typedef vector<RegionID> VectRegID;
typedef VectRegID::iterator VectRegIDIt;

// ----- operator<<

ostream& operator<<(ostream& out, VectPairPoint& vect);
ostream& operator<<(ostream& out, VectPoint& vect);
ostream& operator<<(ostream& out, VectRegID& vect);

/* ----- heap ----- */

typedef vector<PairRegionID> VectPairRegionID;
typedef multiset<PairRegionID, lessPair> PairsRegionIDLess;
typedef PairsRegionIDLess::iterator PairsRegionIDLessIt;
typedef multiset<PairRegionID, lessPairScore> PairsRegionIDLessScore;
typedef PairsRegionIDLessScore::iterator PairsRegionIDLessScoreIt;

/* ----- topk ----- */

typedef priority_queue<PairPoint, VectPairPoint, lessPairScore> TopK;

// --- operator<<

ostream& operator<<(ostream& out, const TopK& topk);

#endif
