/*
  $Id: DataSetUtil.h 4269 2009-02-09 20:08:05Z rares $

  Copyright (C) 2009 by The Regents of the University of California

  Redistribution of this file is permitted under the terms of the
  BSD license

  Date: 09/01/2005
  Author: Rares Vernica <rares (at) ics.uci.edu> 
*/

#ifndef __DataSetUtil_h
#define __DataSetUtil_h

#include <exception>

using namespace std;

#include "DataStruct.h"

// --- const

const int SELECT_LEFT=1;
const int JOIN=2;
const int SELECT_RIGHT=4;
const int SKYLINE_LEFT=8;
const int SKYLINE_RIGHT=16;

// --- dist

inline double dist(const Point& pointA, const unsigned dimA, const Point& pointB, const unsigned long dimB)
{
  return Point(pointA.m_pCoords+dimA, 1).getMinimumDistance(Point(pointB.m_pCoords+dimB, 1));
}

inline double dist(const Region& region, const unsigned dim, const double value) 
{
  double coord[1];
  coord[0]=value;
  return Region(region.m_pLow+dim, region.m_pHigh+dim, 1).getMinimumDistance(Point(coord, 1));
}

inline double dist(const Region& regionA, const unsigned dimA, const Region& regionB, const unsigned long dimB) 
{
  return Region(regionA.m_pLow+dimA,
		regionA.m_pHigh+dimA,
		1).getMinimumDistance(Region(regionB.m_pLow+dimB,
					     regionB.m_pHigh+dimB,
					     1));
}

inline double distMax(const Region& regionA, const unsigned dimA, const Region& regionB, const unsigned long dimB) 
{
  return Region(regionA.m_pLow+dimA,
		regionA.m_pHigh+dimA,
		1).getMaximumDistance(Region(regionB.m_pLow+dimB,
					     regionB.m_pHigh+dimB,
					     1));
}

// --- diff - for dominance check

inline double diff(const double coordP, const double coordQ, const double val)
{
  return abs(coordP-val)-abs(coordQ-val);
}

inline double diff(const double coordP_low, const double coordP_high, const double coordQ, const double val)
{
  return ((coordP_low<=val && coordP_high>=val)?0:min(abs(coordP_low-val), abs(coordP_high-val)))-abs(coordQ-val);
}

inline double diff(const double coordP_low, const double coordP_high, 
		   const double coordQ_low, const double coordQ_high, const double val)
{
  return 
    ((coordP_low<=val && coordP_high>=val)?0:min(abs(coordP_low-val), abs(coordP_high-val)))-
    ((coordQ_low<=val && coordQ_high>=val)?0:max(abs(coordQ_low-val), abs(coordQ_high-val)));
}

/*
  relation:
    R(R0, R1, R2, R3, R4, R5, R6)

  weights:
    W_R_O
    W_R_1
    W_R_2
    W_R_3
    W_R_4
    W_R_5
    W_R_6

  semantics:
    R0 - age
    R1 - wage per hour
    R2 - capital gains
    R3 - capital loses
    R4 - dividends from stocks
    R5 - num persons worked for employer
    R6 - weeks worked in year
*/
// domain min and len

 const double domains[7][2]=
  {{0, 90}, 
   {0, 9999},
   {0, 99999},
   {0, 4608}, 
   {0, 99999},
   {0, 6},
   {0, 2}};
/*
const double domains[7][2]=
  {{0, 1}, 
   {0, 1},
   {0, 1},
   {0, 1}, 
   {0, 1},
   {0, 1},
   {0, 1}};
*/

const unsigned query_no=4;
const unsigned attr_no=7;

const double R[query_no][attr_no]=
  {
    {0, 1200, 0, 0, 6000, 0, 0},
    {0,  500, 0, 0,   90, 0, 0}, // Q1
    // {0,  510, 0, 0,  500, 0, 0},
    {0,  500, 0, 0,  490, 0, 0}, // Q2
    // {0,  590, 0, 0,  100, 0, 0},
    {0,  600, 0, 0,  490, 0, 0}, // Q3
    // {0,  610, 0, 0,  500, 0, 0},
    // {0,  600, 0, 0,  510, 0, 0}
  };

extern unsigned query_crt;
extern double Rn[query_no][attr_no];

const double W_R[attr_no]=
  {0.15, 0.15, 0.1, 0.15, 0.15, 0.1, 0.2};

/*
  query:
    SELECT *
    FROM R
    WHERE R.R0=R.R_0
      AND R.R1=R_1
      AND R.R4=R_4
*/

// ----- Init

void initRn();

// ----- Select&Join

struct Sel_R_0_1st
{
  inline bool operator()(const Point& p)
  {
    return
      abs(p.m_pCoords[0]-Rn[query_crt][0])<numeric_limits<double>::epsilon();
  }

  inline bool operator()(const Region& r)
  {
    return Region(r.m_pLow, r.m_pHigh, 1).containsPoint(Point(Rn[query_crt], 1));
  }
};

struct Sel_R_1_2nd
{
  inline bool operator()(const Point& p)
  {
    return
      abs(p.m_pCoords[1]-Rn[query_crt][1])<numeric_limits<double>::epsilon();
  }

  inline bool operator()(const Region& r)
  {
    return Region(r.m_pLow+1, r.m_pHigh+1, 1).containsPoint(Point(Rn[query_crt]+1, 1));
  }
};

struct Sel_R_4_3rd
{
  inline bool operator()(const Point& p)
  {
    return
      abs(p.m_pCoords[2]-Rn[query_crt][4])<numeric_limits<double>::epsilon();
  }

  inline bool operator()(const Region& r)
  {
    return Region(r.m_pLow+2, r.m_pHigh+2, 1).containsPoint(Point(Rn[query_crt]+4, 1));
  }
};

struct Sel_R_6_3rd
{
  inline bool operator()(const Point& p)
  {
    return
      abs(p.m_pCoords[2]-Rn[query_crt][6])<numeric_limits<double>::epsilon();
  }

  inline bool operator()(const Region& r)
  {
    return Region(r.m_pLow+2, r.m_pHigh+2, 1).containsPoint(Point(Rn[query_crt]+6, 1));
  }
};

struct Join_R_0_1st
{
  inline bool operator()(const PairPoint& pp) 
  {
    return 
      abs(pp.first.m_pCoords[0]-pp.second.m_pCoords[0])<numeric_limits<double>::epsilon();
  }

  inline bool operator()(const PairRegion& pr) 
  {
    return Region(pr.first.m_pLow, pr.first.m_pHigh, 1)
      .intersectsRegion(Region(pr.second.m_pLow, pr.second.m_pHigh, 1));
  }  
};

// ----- Distance

inline double dist_R(const PairRegion pair) 
{
  double d=0;
  d+=dist(pair.first, 0, pair.second, 0);
  d+=dist(pair.first, 1, Rn[query_crt][1]);
  d+=dist(pair.second, 2, Rn[query_crt][4]);
  return d;
}

inline bool dom_R(const PairPoint& ppA, const PairPoint& ppB, const int flag)
{
  bool oneGreater=false;
  double diff_val;

  if (flag&SELECT_LEFT) {
    diff_val=diff(ppB.first.m_pCoords[1], ppA.first.m_pCoords[1], Rn[query_crt][1]);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }
  
  if (flag&JOIN) {
    // R.0=R.0 1st
    diff_val=dist(ppB.first, 0, ppB.second, 0)-dist(ppA.first, 0, ppA.second, 0);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }
  
  if (flag&SELECT_RIGHT) {
    // R.4 3rd
    diff_val=diff(ppB.second.m_pCoords[2], ppA.second.m_pCoords[2], Rn[query_crt][4]);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }
  
  return oneGreater;
}

inline bool dom_R(const PairPoint& pp, const PairRegion& pr, const int flag)
{  
  bool oneGreater=false;
  double diff_val;

  if (flag&SELECT_LEFT) {
    // R.1 2nd
    diff_val=diff(pr.first.m_pLow[1], pr.first.m_pHigh[1], pp.first.m_pCoords[1], Rn[query_crt][1]);    
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }

  if (flag&JOIN) {
    // R.0 1st
    diff_val=dist(pr.first, 0, pr.second, 0)-dist(pp.first, 0, pp.second, 0);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }
  
  if (flag&SELECT_RIGHT) {
    // R.4 3rd
    diff_val=diff(pr.second.m_pLow[2], pr.second.m_pHigh[2], pp.second.m_pCoords[2], Rn[query_crt][4]);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }
  
  return oneGreater;
}

inline bool dom_R(const PairRegion& prA, const PairRegion& prB, const int flag)
{  
  bool oneGreater=false;
  double diff_val;

  if (flag&SELECT_LEFT) {
    // R.1 2nd
    diff_val=diff(prB.first.m_pLow[1], prB.first.m_pHigh[1],
		  prA.first.m_pLow[1], prA.first.m_pHigh[1], Rn[query_crt][1]);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }

  if (flag&JOIN) {
    // R.0 1st
    diff_val=dist(prB.first, 0, prB.second, 0)-distMax(prA.first, 0, prA.second, 0);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }
  
  if (flag&SELECT_RIGHT) {
    // R.4 3rd
    diff_val=diff(prB.second.m_pLow[2], prB.second.m_pHigh[2], 
		  prA.second.m_pLow[2], prA.second.m_pHigh[2], Rn[query_crt][4]);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }
  
  return oneGreater;
}

// --- Score

inline double score_R(const PairPoint& pp)
{
  double sc=0;
  sc+=W_R[0]*abs(pp.first.m_pCoords[0]-pp.second.m_pCoords[0]);
  sc+=W_R[1]*abs(Rn[query_crt][1]-pp.first.m_pCoords[1]);
  sc+=W_R[4]*abs(Rn[query_crt][4]-pp.second.m_pCoords[2]);
  return sc;
}

inline double scoreMin_R(const PairRegion& pr)
{
  double sc=0;
  sc+=W_R[0]*Region(pr.first.m_pLow,
		    pr.first.m_pHigh, 1).getMinimumDistance(Region(pr.second.m_pLow, 
								   pr.second.m_pHigh, 1));
  sc+=W_R[1]*Point(Rn[query_crt]+1, 1).getMinimumDistance(Region(pr.first.m_pLow+1, 
								 pr.first.m_pHigh+1, 1));
  sc+=W_R[4]*Point(Rn[query_crt]+4, 1).getMinimumDistance(Region(pr.second.m_pLow+2, 
								 pr.second.m_pHigh+2, 1));
  return sc;
}

inline double scoreMinMax_R(const PairRegion& pr)
{
  double sc=0;

  unsigned n=3, min_i=n;
  double mindist[n], maxdist[n], weight[n];
  double maxsum, maxmin;
  
  Point p;
  Region r1, r2;

  r1=Region(pr.first.m_pLow, pr.first.m_pHigh, 1);
  r2=Region(pr.second.m_pLow, pr.second.m_pHigh, 1);
  mindist[0]=r1.getMinimumDistance(r2);
  maxdist[0]=r1.getMaximumDistance(r2);
  weight[0]=W_R[0];  

  p=Point(Rn[query_crt]+1, 1);
  r1=Region(pr.first.m_pLow+1, pr.first.m_pHigh+1, 1);
  mindist[1]=r1.getMinimumDistance(p);
  maxdist[1]=r1.getMaximumDistance(p);
  weight[1]=W_R[1];

  p=Point(Rn[query_crt]+4, 1);
  r1=Region(pr.second.m_pLow+2, pr.second.m_pHigh+2, 1);
  mindist[2]=r1.getMinimumDistance(p);
  maxdist[2]=r1.getMaximumDistance(p);
  weight[2]=W_R[4];

  maxsum=0;
  for (unsigned i=0; i<n; i++) {
    maxsum+=maxdist[i];
  }
  
  maxmin=maxsum;

  for (unsigned i=0; i<n; i++) {
    double maxmin_new=maxsum-maxdist[i]+mindist[i];
    if (maxmin_new<maxmin) {
      maxmin=maxmin_new;
      min_i=i;
    }
  }

  for (unsigned i=0; i<n; i++) {
    if (i==min_i) {
      sc+=weight[i]*mindist[i];
    }
    else {
      sc+=weight[i]*maxdist[i];
    }
  }

  return sc;
}

// ----------

// --- cond

inline bool cond(const PairPoint& pp, const int flag=SELECT_LEFT|JOIN|SELECT_RIGHT) 
{
  if (flag&SELECT_LEFT) {
    if (Sel_R_1_2nd()(pp.first)) {
    }
    else {
      return false;
    }
  }
  if (flag&JOIN) {
    if (Join_R_0_1st()(pp)) {
    }
    else {
      return false;
    }
  }
  if (flag&SELECT_RIGHT) {
    if (Sel_R_4_3rd()(pp.second)) {
    }
    else {
      return false;
    }
  }
  return true;
}

inline bool cond(const PairRegion& r, const int flag=SELECT_LEFT|JOIN|SELECT_RIGHT) 
{
  if (flag&SELECT_LEFT) {
    if (Sel_R_1_2nd()(r.first)) {
    }
    else {
      return false;
    }
  }
  if (flag&JOIN) {
    if (Join_R_0_1st()(r)) {
    }
    else {
      return false;
    }
  }
  if (flag&SELECT_RIGHT) {
    if (Sel_R_4_3rd()(r.second)) {
    }
    else {
      return false;
    }
  }
  return true;
}

inline bool cond(const PairRegionID& prID, const int flag=SELECT_LEFT|JOIN|SELECT_RIGHT) 
{
  return cond(PairRegion(prID.first.getRegion(), prID.second.getRegion()), flag);
}

inline bool cond(const PairRegionIDLevel& prIDLevel, const int flag=SELECT_LEFT|JOIN|SELECT_RIGHT) 
{
  return cond(static_cast<PairRegionID>(prIDLevel), flag);
}

// --- dist

inline double dist(const PairRegion& pair) 
{
  return dist_R(pair);
}

inline double dist(const PairRegionID& pair)
{
  return dist(PairRegion(pair.first.getRegion(), pair.second.getRegion()));
}

// --- dom

inline bool dom(const PairPoint& ppA, const PairPoint& ppB, const int flag=SELECT_LEFT|JOIN|SELECT_RIGHT)
{
  return dom_R(ppA, ppB, flag);
}

inline bool dom(const PairPoint& pp, const PairRegion& pr, const int flag=SELECT_LEFT|JOIN|SELECT_RIGHT)
{
  return dom_R(pp, pr, flag);
}

inline bool dom(const PairRegion& prA, const PairRegion& prB, const int flag=SELECT_LEFT|JOIN|SELECT_RIGHT) 
{
  return dom_R(prA, prB, flag);
}

inline bool dom(const PairPoint& pp, const PairRegionID& pr, const int flag=SELECT_LEFT|JOIN|SELECT_RIGHT) 
{
  return dom(pp, PairRegion(pr.first.getRegion(), pr.second.getRegion()));
}

inline bool dom(const PairRegionID& prA, const PairRegionID& prB, const int flag=SELECT_LEFT|JOIN|SELECT_RIGHT)
{
  return dom_R(PairRegion(prA.first.getRegion(), prA.second.getRegion()), 
	       PairRegion(prB.first.getRegion(), prB.second.getRegion()), 
	       flag);
}

// --- score

inline double scoreMin(const PairPoint& pp)
{
  return score_R(pp);
}

inline double scoreMinMax(const PairPoint& pp)
{
  return score_R(pp);
}

inline double scoreMin(const PairRegion& pr)
{
  return scoreMin_R(pr);
}

inline double scoreMinMax(const PairRegion& pr)
{
  return scoreMinMax_R(pr);
}

inline double scoreMin(const PairRegionID& prID) 
{
  return scoreMin(PairRegion(prID.first.getRegion(), prID.second.getRegion()));
}

inline double scoreMinMax(const PairRegionID& prID) 
{
  return scoreMinMax(PairRegion(prID.first.getRegion(), prID.second.getRegion()));
}

// ----------------------------------------------------------------------------------------
// ----- OLD -----
// ----------------------------------------------------------------------------------------

// ----- Select&Join
/*
struct Sel_R_01
{
  inline bool operator()(const Point& p)
  {
    return
      abs(p.m_pCoords[0]-R_0)<numeric_limits<double>::epsilon() && 
      abs(p.m_pCoords[1]-R_1)<numeric_limits<double>::epsilon();
  }

  inline bool operator()(const Region& r)
  {
    return Region(r.m_pLow, r.m_pHigh, 2).containsPoint(Point(R_01, 2));
  }
};

struct Sel_R_46
{
  inline bool operator()(const Point& p)
  {
    return 
      abs(p.m_pCoords[4]-R_4)<numeric_limits<double>::epsilon() && 
      abs(p.m_pCoords[6]-R_6)<numeric_limits<double>::epsilon();
  }

  inline bool operator()(const Region& r)
  {
    double low[2], high[2];
    low[0]=r.m_pLow[4];
    low[1]=r.m_pLow[6];
    high[0]=r.m_pHigh[4];
    high[1]=r.m_pHigh[6];

    return Region(low, high, 2).containsPoint(Point(R_46, 2));
  }
};

struct Sel_R_45
{
  inline bool operator()(const Point& p)
  {
    return 
      abs(p.m_pCoords[4]-R_4)<numeric_limits<double>::epsilon() && 
      abs(p.m_pCoords[5]-R_5)<numeric_limits<double>::epsilon();
  }

  inline bool operator()(const Region& r)
  {
    double low[2], high[2];
    low[0]=r.m_pLow[4];
    low[1]=r.m_pLow[5];
    high[0]=r.m_pHigh[4];
    high[1]=r.m_pHigh[5];

    return Region(low, high, 2).containsPoint(Point(R_45, 2));
  }
};

struct Join_R_23
{
  inline bool operator()(const PairPoint& pp) 
  {
    return 
      abs(pp.first.m_pCoords[2]-pp.second.m_pCoords[2])<numeric_limits<double>::epsilon() && 
      abs(pp.first.m_pCoords[3]-pp.second.m_pCoords[3])<numeric_limits<double>::epsilon();
  }

  inline bool operator()(const PairRegion& pr) 
  {
    return Region(pr.first.m_pLow+2, pr.first.m_pHigh+2, 2)
      .intersectsRegion(Region(pr.second.m_pLow+2, pr.second.m_pHigh+2, 2));
  }  
};
*/
// ----- Dominance 
/*
struct Dom_R_01
{
  bool operator()(const Point& x, const Region& y) 
  {
    bool oneGreater=false;
    double diff_val;
    // R.0
    diff_val=diff(y.m_pLow[0], y.m_pHigh[0], x.m_pCoords[0], R_0);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    // R.1
    diff_val=diff(y.m_pLow[1], y.m_pHigh[1], x.m_pCoords[1], R_1);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    return oneGreater;
  }    

  bool operator()(const Point& x, const Point& y) 
  {
    bool oneGreater=false;
    double diff_val;
    // R.0
    diff_val=diff(y.m_pCoords[0], x.m_pCoords[0], R_0);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    // R.1
    diff_val=diff(y.m_pCoords[1], x.m_pCoords[1], R_1);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    return oneGreater;
  }    
};

struct Dom_R_46
{
  bool operator()(const Point& x, const Region& y) 
  {
    bool oneGreater=false;
    double diff_val;
    // R.4
    diff_val=diff(y.m_pLow[4], y.m_pHigh[4], x.m_pCoords[4], R_4);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    // R.6
    diff_val=diff(y.m_pLow[6], y.m_pHigh[6], x.m_pCoords[6], R_6);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    return oneGreater;
  }

  bool operator()(const Point& x, const Point& y) 
  {
    bool oneGreater=false;
    double diff_val;
    // R.4
    diff_val=diff(y.m_pCoords[4], x.m_pCoords[4], R_4);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    // R.6
    diff_val=diff(y.m_pCoords[6], x.m_pCoords[6], R_6);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    return oneGreater;
  }    
};
*/
// --- Compare
/*
struct greaterRegionID_R_01 {
  bool operator()(const RegionID& x, const RegionID& y) 
  {
    Point p=Point(R_01, 2);
    return Region(x.reg.m_pLow, x.reg.m_pHigh, 2).getMinimumDistance(p)>
      Region(y.reg.m_pLow, y.reg.m_pHigh, 2).getMinimumDistance(p);
  }
};

struct greaterRegionID_R_46 {
  bool operator()(const RegionID& x, const RegionID& y) 
  {
    Point p=Point(R_46, 2);

    double lowX[2], highX[2], lowY[2], highY[2];
    lowX[0]=x.reg.m_pLow[4];
    lowX[1]=x.reg.m_pLow[6];
    highX[0]=x.reg.m_pHigh[4];
    highX[1]=x.reg.m_pHigh[6];
    lowY[0]=y.reg.m_pLow[4];
    lowY[1]=y.reg.m_pLow[6];
    highY[0]=y.reg.m_pHigh[4];
    highY[1]=y.reg.m_pHigh[6];

    return Region(lowX, highX, 2).getMinimumDistance(p)>Region(lowY, highY, 2).getMinimumDistance(p);
  }
};
*/
/*
  relations:
    R(A, B, C, D), S(C, D, E, F)

  query:
    SELECT *
    FROM R, S
    WHERE R.A=R_A
      AND R.B=R_B
      AND R.C=S.C
      AND R.D=S.D
      AND S.E=S_E
      AND S.F=S_F

  weights:
    W_R_A
    W_R_B
    W_RS_C
    W_RS_D
    W_S_E
    W_S_F
*/
/*
const double R_A=0.5;
const double R_B=0.5;
const double S_E=0.5;
const double S_F=0.5;
const double Eps=0.1;

const double W_R_A=0.15;
const double W_R_B=0.15;
const double W_RS_C=0.2;
const double W_RS_D=0.15;
const double W_S_E=0.15;
const double W_S_F=0.2;

const double R_AB[2]=
  {
    R_A, R_B
  };

const double S_EF[2]=
  {
    S_E, S_F
  };

const double R_ABLow[2]=
  {
    R_AB[0]-Eps, R_AB[1]-Eps
  };
const double R_ABHigh[2]=
  {
    R_AB[0]+Eps, R_AB[1]+Eps
  };
const double S_EFLow[2]=
  {
    S_EF[0]-Eps, S_EF[1]-Eps
  };
const double S_EFHigh[2]=
  {
    S_EF[0]+Eps, S_EF[1]+Eps
  };

// ----- Select&Join

struct Sel_R_AB
{
  bool operator()(const Point& p)
  {
    if (Point(p.m_pCoords, 2)
	.intersectsShape(Region(R_ABLow, R_ABHigh, 2))
	) {
    }
    else {
      return false;
    }
    return true;
  }

  bool operator()(const Region& r)
  {
    if (Region(r.m_pLow, r.m_pHigh, 2)
	.intersectsRegion(Region(R_ABLow, R_ABHigh, 2))
	) {
    }
    else {
      return false;
    }
    return true;
  }
};

struct Join_RS_CD
{
  bool operator()(const PairPoint& pp) 
  {
    const double second_coordLow[2]=
      {
	pp.second.m_pCoords[0]-Eps, pp.second.m_pCoords[1]-Eps
      };
    const double second_coordHigh[2]=
      {
	pp.second.m_pCoords[0]+Eps, pp.second.m_pCoords[1]+Eps
      };

    if (Point(pp.first.m_pCoords+2, 2)
	.intersectsShape(Region(second_coordLow, second_coordHigh, 2))
	) {
    }
    else {
      return false;
    }
  }

  bool operator()(const PairRegion& pr) 
  {
    const double second_coordLow[2]=
      {
	pr.second.m_pLow[0]-Eps, pr.second.m_pLow[1]-Eps
      };
    const double second_coordHigh[2]=
      {
	pr.second.m_pHigh[0]+Eps, pr.second.m_pHigh[1]+Eps
      };

    if (Region(pr.first.m_pLow+2, pr.first.m_pHigh+2, 2)
	.intersectsRegion(Region(second_coordLow, second_coordHigh, 2))
// 	.intersectsRegion(Region(pr.second.m_pLow, pr.second.m_pHigh, 2))
	) {
    }
    else {
      return false;
    }
  }  
};

struct Sel_S_EF
{
  bool operator()(const Point& p)
  {
    if (Point(p.m_pCoords+2, 2)
	.intersectsShape(Region(S_EFLow, S_EFHigh, 2))
	) {
    }
    else {
      return false;
    }
    return true;
  }

  bool operator()(const Region& r)
  {
    if (Region(r.m_pLow+2, r.m_pHigh+2, 2)
	.intersectsRegion(Region(S_EFLow, S_EFHigh, 2))
	) {
    }
    else {
      return false;
    }
    return true;
  }
};

// ----- Dominance

inline bool dom_RS(const PairPoint& ppA, const PairPoint& ppB, const int flag=SELECT_LEFT|JOIN|SELECT_RIGHT)
{
  bool oneGreater=false;
  double diff_val;

  if (flag&SELECT_LEFT) {
    // R.A
    diff_val=diff(ppB.first.m_pCoords[0], ppA.first.m_pCoords[0], R_A);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    // R.B
    diff_val=diff(ppB.first.m_pCoords[1], ppA.first.m_pCoords[1], R_B);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }
  
  if (flag&JOIN) {
    // R.C=S.C
    diff_val=(dist(ppB.first, 2, ppB.second, 0)-
	  dist(ppA.first, 2, ppA.second, 0));
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    // R.D=S.D
    diff_val=(dist(ppB.first, 3, ppB.second, 1)-
	  dist(ppA.first, 3, ppA.second, 1));
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }
  
  if (flag&SELECT_RIGHT) {
    // S.E
    diff_val=diff(ppB.second.m_pCoords[2], ppA.second.m_pCoords[2], S_E);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    // S.F
    diff_val=diff(ppB.second.m_pCoords[3], ppA.second.m_pCoords[3], S_F);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }
  
  return oneGreater;
}

inline bool dom_RS(const PairPoint& pp, const PairRegion& pr, const int flag=SELECT_LEFT|JOIN|SELECT_RIGHT)
{
  bool oneGreater=false;
  double diff_val;

  if (flag&SELECT_LEFT) {
    // R.A
    diff_val=diff(pr.first.m_pLow[0], pr.first.m_pHigh[0], pp.first.m_pCoords[0], R_A);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    // R.B
    diff_val=diff(pr.first.m_pLow[1], pr.first.m_pHigh[1], pp.first.m_pCoords[1], R_B);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }

  if (flag&JOIN) {
    // R.C=S.C
    diff_val=(dist(pr.first, 2, pr.second, 0)-
	  dist(pp.first, 2, pp.second, 0));
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    // R.D=S.D
    diff_val=(dist(pr.first, 3, pr.second, 1)-
	  dist(pp.first, 3, pp.second, 1));
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }
  
  if (flag&SELECT_RIGHT) {
    // S.E
    diff_val=diff(pr.second.m_pLow[2], pr.second.m_pHigh[2], pp.second.m_pCoords[2], S_E);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    // S.F
    diff_val=diff(pr.second.m_pLow[3], pr.second.m_pHigh[3], pp.second.m_pCoords[3], S_F);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
  }
  
  return oneGreater;
}

struct Dom_R_AB 
{
  bool operator()(const Point& x, const Region& y) 
  {
    bool oneGreater=false;
    double diff_val;
    // R.A
    diff_val=diff(y.m_pLow[0], y.m_pHigh[0], x.m_pCoords[0], R_A);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    // R.B
    diff_val=diff(y.m_pLow[1], y.m_pHigh[1], x.m_pCoords[1], R_B);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    return oneGreater;
  }    

  bool operator()(const Point& x, const Point& y) 
  {
    bool oneGreater=false;
    double diff_val;
    // R.A
    diff_val=diff(y.m_pCoords[0], x.m_pCoords[0], R_A);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    // R.B
    diff_val=diff(y.m_pCoords[1], x.m_pCoords[1], R_B);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    return oneGreater;
  }    
};

struct Dom_S_EF 
{
  bool operator()(const Point& x, const Region& y) 
  {
    bool oneGreater=false;
    double diff_val;
    // S.E
    diff_val=diff(y.m_pLow[2], y.m_pHigh[2], x.m_pCoords[2], S_E);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    // S.F
    diff_val=diff(y.m_pLow[3], y.m_pHigh[3], x.m_pCoords[3], S_F);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    return oneGreater;
  }

  bool operator()(const Point& x, const Point& y) 
  {
    bool oneGreater=false;
    double diff_val;
    // S.E
    diff_val=diff(y.m_pCoords[2], x.m_pCoords[2], S_E);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    // S.F
    diff_val=diff(y.m_pCoords[3], x.m_pCoords[3], S_F);
    if (diff_val>numeric_limits<double>::epsilon()) {
      oneGreater=true;
    }
    else if (diff_val<-numeric_limits<double>::epsilon()) {
      return false;
    }
    return oneGreater;
  }    
};
*/
// --- Compare
/*
struct greaterRegionID_R_AB {
  bool operator()(const RegionID& x, const RegionID& y) 
  {
    Point p=Point(R_AB, 2);
    return Region(x.reg.m_pLow, x.reg.m_pHigh, 2).getMinimumDistance(p)>
      Region(y.reg.m_pLow, y.reg.m_pHigh, 2).getMinimumDistance(p);
  }
};

struct greaterRegionID_S_EF {
  bool operator()(const RegionID& x, const RegionID& y) 
  {
    Point p=Point(S_EF, 2);
    return Region(x.reg.m_pLow+2, x.reg.m_pHigh+2, 2).getMinimumDistance(p)>
      Region(y.reg.m_pLow+2, y.reg.m_pHigh+2, 2).getMinimumDistance(p);
  }
};
*/
// --- Score
 /*
inline double score_RS(const PairPoint& pp)
{
  double sc=0;
  sc+=W_R_A*abs(R_A-pp.first.m_pCoords[0]);
  sc+=W_R_B*abs(R_B-pp.first.m_pCoords[1]);
  sc+=W_RS_C*abs(pp.first.m_pCoords[2]-pp.second.m_pCoords[0]);
  sc+=W_RS_D*abs(pp.first.m_pCoords[3]-pp.second.m_pCoords[1]);
  sc+=W_S_E*abs(S_E-pp.second.m_pCoords[2]);
  sc+=W_S_F*abs(S_F-pp.second.m_pCoords[3]);
  return sc;
}

inline double score_RS(const PairRegion& pr)
{
  double sc=0;
  sc+=W_R_A*Point(R_AB, 1).getMinimumDistance(Region(pr.first.m_pLow, 
						     pr.first.m_pHigh, 1));
  sc+=W_R_B*Point(R_AB+1, 1).getMinimumDistance(Region(pr.first.m_pLow+1, 
						       pr.first.m_pHigh+1, 1));
  sc+=W_RS_C*Region(pr.first.m_pLow+2,
		    pr.first.m_pHigh+2, 1).getMinimumDistance(Region(pr.second.m_pLow, 
								     pr.second.m_pHigh, 1));
  sc+=W_RS_D*Region(pr.first.m_pLow+3,
		    pr.first.m_pHigh+3, 1).getMinimumDistance(Region(pr.second.m_pLow+1, 
								     pr.second.m_pHigh+1, 1));
  sc+=W_S_E*Point(S_EF, 1).getMinimumDistance(Region(pr.second.m_pLow+2, 
						     pr.second.m_pHigh+2, 1));
  sc+=W_S_F*Point(S_EF+1, 1).getMinimumDistance(Region(pr.second.m_pLow+3, 
						       pr.second.m_pHigh+3, 1));
  return sc;
}
*/

#endif

