/*
  $Id: DataStruct.h 4269 2009-02-09 20:08:05Z rares $

  Copyright (C) 2009 by The Regents of the University of California

  Redistribution of this file is permitted under the terms of the
  BSD license

  Date: 09/01/2005
  Author: Rares Vernica <rares (at) ics.uci.edu> 
*/

#ifndef __DataStruct_h
#define __DataStruct_h

#include <iostream>
#include <functional>

using namespace std;

#include "RTreeUtil.h"

/*
 * Data Structures
 * ---
 * - heap
 *   - of pairs
 *   - sorted, pair with the minimum relaxation first
 *   - insert
 *   - pop
 * - pair
 *   - of regions or data points
 * - topk
 *   - priority queue of pair poins
 * 
 */

extern unsigned global_timeIO;

/* ----- pair ----- */

class RegionID 
{
public:
  unsigned long id, pid;
  bool index;  
private:
  RTree::RTree *rtree;

public:
  RegionID() {}
  RegionID(unsigned long i, bool idx, RTree::RTree *r): id(i), pid(0), index(idx), rtree(r) {}
  RegionID(unsigned long i, unsigned long pi, RTree::RTree *r): id(i), pid(pi), index(false), rtree(r) {}

  const Region getRegion() const;

  friend ostream& operator<<(ostream& out, const RegionID& regid);
};

inline const Region RegionID::getRegion() const 
{
  if (index) {
    return rtree->readNode(id)->m_nodeMBR;
  }
  else {
    return *(rtree->readNode(pid)->m_ptrMBR[id]);
  }
}

class RegionIDLevel: public RegionID 
{
public:
  unsigned level;
  
  RegionIDLevel() {}
  RegionIDLevel(unsigned long i, bool idx, RTree::RTree *r, unsigned l): RegionID(i, idx, r), level(l) {}
  RegionIDLevel(unsigned long i, unsigned long pi, RTree::RTree *r, unsigned l): RegionID(i, pi, r), level(l) {}

  friend ostream& operator<<(ostream& out, const RegionIDLevel& regidlevel);
};

typedef pair<Point, Point> PairPoint;
typedef pair<Region, Region> PairRegion;
typedef pair<RegionID, RegionID> PairRegionID;
typedef pair<RegionIDLevel, RegionIDLevel> PairRegionIDLevel;

// --- operator<<

ostream& operator<<(ostream& out, const PairRegion& pair);
ostream& operator<<(ostream& out, const PairPoint& pair);
ostream& operator<<(ostream& out, const RegionID& regid);
ostream& operator<<(ostream& out, const RegionIDLevel& regidlevel);
ostream& operator<<(ostream& out, const PairRegionID& pair);
ostream& operator<<(ostream& out, const PairRegionIDLevel& pair);

#endif
