/*
  $Id: RTreeUtil.cpp 4269 2009-02-09 20:08:05Z rares $

  Copyright (C) 2009 by The Regents of the University of California

  Redistribution of this file is permitted under the terms of the
  BSD license

  Date: 09/01/2005
  Author: Rares Vernica <rares (at) ics.uci.edu> 
*/

#include <iostream>

using namespace std;

#include "DataSetUtil.h"

const double fillFactor=0.7;
// const unsigned indexCapacity=200;
// const unsigned leafCapacity=200;

/* ----- rtree ----- */
void createRTree(string input_file,
		 string output_file,
		 RTree::RTree*& rtree,
		 IStorageManager*& storagemngr,
		 StorageManager::IBuffer*& file,
		 unsigned index_capacity,
		 unsigned no_attr,
		 string attr_map,
		 unsigned no_rec)
{
  cerr<<"createRTree "<<input_file<<" "<<output_file<<endl;
  try {
    ifstream fin(input_file.c_str());
    if (!fin) {
      cerr<<"Cannot open file "<<input_file<<"."<<endl;
    }

    if (output_file.length()==0) {
      storagemngr=StorageManager::createNewMemoryStorageManager();
    }
    else {
      storagemngr=StorageManager::createNewDiskStorageManager(output_file, 4096);
    }
    file=StorageManager::createNewRandomEvictionsBuffer(*storagemngr, 10, false);

    long long indexIdentifier;
    unsigned no_attr_real=0;

    for (unsigned i=0; i<attr_map.size(); i++) {
      if (attr_map[i]=='1')
	no_attr_real++;
    }

    rtree=dynamic_cast<RTree::RTree*>(RTree::createNewRTree(*file,
							    fillFactor,
							    index_capacity,
							    index_capacity, 
							    no_attr_real,
							    SpatialIndex::RTree::RV_RSTAR, 
							    indexIdentifier));

    unsigned long id=0;
    double x;
    double plow[no_attr_real], phigh[no_attr_real];
    Region r;
    Tools::ResourceUsage ru;

    ru.start();
    while (fin) {
      for (unsigned i=0, j=0; i<no_attr; i++) {
	fin>>x;
	if (!fin.good()) continue;
	if (attr_map[i]=='1') {
	  plow[j]=phigh[j]=(x-domains[i][0])/domains[i][1];
	  j++;
	}
      }

      r=Region(plow, phigh, no_attr_real);
      
      rtree->insertData(0, 0, r, id++);

      if (id==no_rec) {
	break;
      }
    }
    ru.stop();

    bool ret=rtree->isIndexValid();
    if (ret==false) cerr<<"ERROR: Structure is invalid!"<<endl;
    else cerr<<"The structure seems O.K."<<endl;

//     cerr<<*(rtree->readNode(rtree->m_rootID))<<endl;
//     cerr<<(unsigned)rtree<<endl;
  }
  catch (Tools::Exception& e) {
    cerr<<"******ERROR******"<<endl;
    std::string s=e.what();
    cerr<<s<<endl;
  }
  catch (...) {
    cerr<<"******ERROR******"<<endl;
    cerr<<"other exception"<<endl;
  }
}

void loadRTree(string input_file,
	       RTree::RTree*& rtree,
	       IStorageManager*& storagemngr,
	       StorageManager::IBuffer*& file)
{
  cerr<<"loadRTree "<<input_file<<endl;
  try {
    storagemngr=StorageManager::loadDiskStorageManager(input_file);
    file=StorageManager::createNewRandomEvictionsBuffer(*storagemngr, 10, false);
    rtree=dynamic_cast<RTree::RTree*>(RTree::loadRTree(*file, 1));
  }
  catch (Tools::Exception& e) {
    cerr<<"******ERROR******"<<endl;
    std::string s=e.what();
    cerr<<s<<endl;
  }
  catch (...) {
    cerr<<"******ERROR******"<<endl;
    cerr<<"other exception"<<endl;
  }
}

ostream& operator<<(ostream& out, const INode& node) 
{
  out<<"IEntry.getIdentifier\t"<<node.getIdentifier()<<endl;
  out<<"INode.getLevel\t"<<node.getLevel()<<endl;
  out<<"is"<<(node.isLeaf()?"Leaf":"Index")<<endl;
  out<<"INode.getChildrenCount\t"<<node.getChildrenCount()<<endl;
  for (unsigned long cChild=0; cChild<node.getChildrenCount(); cChild++) {
    out<<"\tgetChildIdentifier\t"<<node.getChildIdentifier(cChild)<<endl;
  }
  return out;
}

void OutputReadableQueryStrategy::getNextEntry(const IEntry& entry, long& nextEntry, bool& hasNext)
{
  IShape *ps;
  entry.getShape(&ps);
  Region* pr=dynamic_cast<Region*>(ps);
    
  const INode* n=dynamic_cast<const INode*>(&entry);

  cout<<*n<<*pr<<endl<<endl;
    
  delete ps;

  if (n!=0 && n->getLevel()>0) {
    for (unsigned long cChild=0; cChild<n->getChildrenCount(); cChild++) {
      ids.push(n->getChildIdentifier(cChild));
    }
  }
    
  if (!ids.empty()) {
    nextEntry=ids.front();
    ids.pop();
    hasNext=true;
  }
  else {
    hasNext=false;
  }
}

void OutputGnuplotQueryStrategy::getNextEntry(const IEntry& entry, long& nextEntry, bool& hasNext)
{
  IShape *ps;
  entry.getShape(&ps);
  Region* pr=dynamic_cast<Region*>(ps);
    
  cout<<pr->m_pLow[0]<<" "<<pr->m_pLow[1]<<endl;
  cout<<pr->m_pHigh[0]<<" "<<pr->m_pLow[1]<<endl;
  cout<<pr->m_pHigh[0]<<" "<<pr->m_pHigh[1]<<endl;
  cout<<pr->m_pLow[0]<<" "<<pr->m_pHigh[1]<<endl;
  cout<<pr->m_pLow[0]<<" "<<pr->m_pLow[1]<<endl<<endl<<endl;

  delete ps;
    
  const INode* n=dynamic_cast<const INode*>(&entry);
    
  if (n!=0 && n->getLevel()>0) {
    for (unsigned long cChild=0; cChild<n->getChildrenCount(); cChild++) {
      ids.push(n->getChildIdentifier(cChild));
    }
  }
    
  if (!ids.empty()) {
    nextEntry=ids.front();
    ids.pop();
    hasNext=true;
  }
  else {
    hasNext=false;
  }
}
