/*
  $Id: Relax.cpp 4269 2009-02-09 20:08:05Z rares $

  Copyright (C) 2009 by The Regents of the University of California

  Redistribution of this file is permitted under the terms of the
  BSD license

  Date: 09/01/2005
  Author: Rares Vernica <rares (at) ics.uci.edu> 
*/

#include <functional>
#include <iomanip>
#include <iterator>
#include <utility>

using namespace std;

#include "RTreeUtil.h"
#include "DataStruct.h"
#include "DataSetUtil.h"
#include "DataCont.h"
#include "Skyline.h"

// ---------- SKYLINE

/*
 * 1 Skyline
 * 
 * - returns skyline
 *
 */
void Relax_Skyline(RTree::RTree& rtreeR, RTree::RTree& rtreeS, VectPairPoint& skyline, int flag_cond, int flag_dom) 
{
  PairsRegionIDLess *heap=new PairsRegionIDLess();
//   VectPairRegionID *heap=new VectPairRegionID();
  VectRegID
    *vectRegIDR=new VectRegID(),
    *vectRegIDS=new VectRegID();

  PairRegionID pair=PairRegionID(RegionID(rtreeR.m_rootID, true, &rtreeR),
				 RegionID(rtreeS.m_rootID, true, &rtreeS));
  global_timeIO+=2;
  
  if (flag_cond?cond(pair, flag_cond):1) { // condition should be verified (select or join)
    heap->insert(pair);
//     heap->push_back(pair);
  }
  while(!heap->empty()) {
    pair=*heap->begin();
    heap->erase(heap->begin());
//     pair=heap->back();
//     heap->pop_back(); 
    if (flag_dom?!dom(skyline, pair, flag_dom):1) { // pair should not be already dominated
      if (pair.first.index || pair.second.index) { // if any of the pair elements is a region
	if (pair.first.index) {
	  global_timeIO+=1;
	  RTree::NodePtr node=rtreeR.readNode(pair.first.id);
	  for (unsigned long cChild=0; cChild<node->getChildrenCount(); cChild++) {
	    if (node->isIndex()) {
	      vectRegIDR->push_back(RegionID(node->getChildIdentifier(cChild), true, &rtreeR));
	    }
	    else {
	      vectRegIDR->push_back(RegionID(cChild, pair.first.id, &rtreeR));
	    }
	  }
	}
	else {
	  vectRegIDR->push_back(pair.first);
	}
	if (pair.second.index) {
	  global_timeIO+=1;
	  RTree::NodePtr node=rtreeS.readNode(pair.second.id);
	  for (unsigned long cChild=0; cChild<node->getChildrenCount(); cChild++) {
	    if (node->isIndex()) {
	      vectRegIDS->push_back(RegionID(node->getChildIdentifier(cChild), true, &rtreeS));
	    }
	    else {
	      vectRegIDS->push_back(RegionID(cChild, pair.second.id, &rtreeS));
	    }
	  }
	}
	else {
	  vectRegIDS->push_back(pair.second);
	}
	for (VectRegIDIt itR=vectRegIDR->begin(); itR!=vectRegIDR->end(); itR++) {
	  for (VectRegIDIt itS=vectRegIDS->begin(); itS!=vectRegIDS->end(); itS++) {
	    PairRegionID pair=PairRegionID(*itR, *itS);
	    if ((flag_cond?cond(pair, flag_cond):1) &&
		(flag_dom?!dom(skyline, pair, flag_dom):1)) {
// 	      if (flag_dom) {
// 		removeDom(*heap, pair, flag_dom);
// 	      }
	      heap->insert(pair);
// 	      heap->push_back(pair);
	    }
	  }
	}
	vectRegIDR->clear();
	vectRegIDS->clear();
      }
      else { // if pair of points
	PairPoint pairPoint=PairPoint(Point(pair.first.getRegion().m_pLow, pair.first.getRegion().m_dimension),
				      Point(pair.second.getRegion().m_pHigh, pair.second.getRegion().m_dimension));
// 	if (flag_dom) {
// 	  removeDom(skyline, pairPoint, flag_dom);
//  	  removeDom(*heap, pairPoint, flag_dom);
// 	}
	skyline.push_back(pairPoint);
      }
    }
  }

  delete vectRegIDR;
  delete vectRegIDS;
  delete heap;
}

// ---------- TOPK

/*
 * 2 TopK
 *
 * - returns top-k
 * 
 */
void Relax_TopK(RTree::RTree& rtreeR, RTree::RTree& rtreeS, VectPairPoint& topk, unsigned k, int flag_cond) 
{
  if (k==0) {
    return;
  }
  PairsRegionIDLessScore *heap=new PairsRegionIDLessScore();
  VectRegID
    *vectRegIDR=new VectRegID(),
    *vectRegIDS=new VectRegID();

  PairRegionID pair=PairRegionID(RegionID(rtreeR.m_rootID, true, &rtreeR),
				 RegionID(rtreeS.m_rootID, true, &rtreeS));
  global_timeIO+=2;

  if (flag_cond?cond(pair, flag_cond):1) {
    // condition should be verified (select or join)
    heap->insert(pair);
  }
  while(!heap->empty()) {
    pair=*heap->begin();
    heap->erase(heap->begin());
    if (pair.first.index || pair.second.index) { // if any of the pair elements is a region
      if (pair.first.index) {
	global_timeIO+=1;
	RTree::NodePtr node=rtreeR.readNode(pair.first.id);
	for (unsigned long cChild=0; cChild<node->getChildrenCount(); cChild++) {
	  if (node->isIndex()) {
	    vectRegIDR->push_back(RegionID(node->getChildIdentifier(cChild), true, &rtreeR));
	  }
	  else {
	    vectRegIDR->push_back(RegionID(cChild, pair.first.id, &rtreeR));
	  }
	}
      }
      else {
	vectRegIDR->push_back(pair.first);
      }
      if (pair.second.index) {
	global_timeIO+=1;
	RTree::NodePtr node=rtreeS.readNode(pair.second.id);
	for (unsigned long cChild=0; cChild<node->getChildrenCount(); cChild++) {
	  if (node->isIndex()) {
	    vectRegIDS->push_back(RegionID(node->getChildIdentifier(cChild), true, &rtreeS));
	  }
	  else {
	    vectRegIDS->push_back(RegionID(cChild, pair.second.id, &rtreeS));
	  }
	}
      }
      else {
	vectRegIDS->push_back(pair.second);
      }
      for (VectRegIDIt itR=vectRegIDR->begin(); itR!=vectRegIDR->end(); itR++) {
	for (VectRegIDIt itS=vectRegIDS->begin(); itS!=vectRegIDS->end(); itS++) {
	  PairRegionID pair=PairRegionID(*itR, *itS);
	  if ((flag_cond?cond(pair, flag_cond):1) &&
	      (heap->size()<k-topk.size() || scoreMin(pair)<scoreMinMax(*heap->rbegin()))) {
	    heap->insert(pair);
	  }
	}
      }
      vectRegIDR->clear();
      vectRegIDS->clear();
    }
    else {// if pair of points
      PairPoint pairPoint=PairPoint(Point(pair.first.getRegion().m_pLow, pair.first.getRegion().m_dimension),
				    Point(pair.second.getRegion().m_pHigh, pair.second.getRegion().m_dimension));
      topk.push_back(pairPoint);
      if (topk.size()==k) {
	break;
      }
    }
    prune(heap, k-topk.size());
  }
  
  delete vectRegIDR;
  delete vectRegIDS;  
  delete heap;  
}

// ---------- TOPK over SKYLINE

/*
 * 3 TopK over Skyline
 *
 * - returns top-k over skyline
 * 
 */
void Relax_TopK(RTree::RTree& rtreeR, RTree::RTree& rtreeS, VectPairPoint& topk, unsigned k, int flag_cond, int flag_dom) 
{
  if (k==0) {
    return;
  }
  PairsRegionIDLessScore *heap=new PairsRegionIDLessScore();
  VectRegID
    *vectRegIDR=new VectRegID(),
    *vectRegIDS=new VectRegID();

  PairRegionID pair=PairRegionID(RegionID(rtreeR.m_rootID, true, &rtreeR),
				 RegionID(rtreeS.m_rootID, true, &rtreeS));
  global_timeIO+=2;

  if (flag_cond?cond(pair, flag_cond):1) {
    // condition should be verified (select or join)
    heap->insert(pair);
  }
  while(!heap->empty()) {
    pair=*heap->begin();
    heap->erase(heap->begin());
    if (flag_dom?!dom(topk, pair, flag_dom):1) { // pair should not be already dominated
      if (pair.first.index || pair.second.index) { // if any of the pair elements is a region
	if (pair.first.index) {
	  global_timeIO+=1;
	  RTree::NodePtr node=rtreeR.readNode(pair.first.id);
	  for (unsigned long cChild=0; cChild<node->getChildrenCount(); cChild++) {
	    if (node->isIndex()) {
	      vectRegIDR->push_back(RegionID(node->getChildIdentifier(cChild), true, &rtreeR));
	    }
	    else {
	      vectRegIDR->push_back(RegionID(cChild, pair.first.id, &rtreeR));
	    }
	  }
	}
	else {
	  vectRegIDR->push_back(pair.first);
	}
	if (pair.second.index) {
	  global_timeIO+=1;
	  RTree::NodePtr node=rtreeS.readNode(pair.second.id);
	  for (unsigned long cChild=0; cChild<node->getChildrenCount(); cChild++) {
	    if (node->isIndex()) {
	      vectRegIDS->push_back(RegionID(node->getChildIdentifier(cChild), true, &rtreeS));
	    }
	    else {
	      vectRegIDS->push_back(RegionID(cChild, pair.second.id, &rtreeS));
	    }
	  }
	}
	else {
	  vectRegIDS->push_back(pair.second);
	}
	for (VectRegIDIt itR=vectRegIDR->begin(); itR!=vectRegIDR->end(); itR++) {
	  for (VectRegIDIt itS=vectRegIDS->begin(); itS!=vectRegIDS->end(); itS++) {
	    PairRegionID pair=PairRegionID(*itR, *itS);
	    if ((flag_cond?cond(pair, flag_cond):1) &&
		(flag_dom?!dom(topk, pair, flag_dom):1)) {
	      heap->insert(pair);
	    }
	  }
	}
	vectRegIDR->clear();
	vectRegIDS->clear();
      }
      else { // if pair of points
	PairPoint pairPoint=PairPoint(Point(pair.first.getRegion().m_pLow, pair.first.getRegion().m_dimension),
				      Point(pair.second.getRegion().m_pHigh, pair.second.getRegion().m_dimension));
	topk.push_back(pairPoint);
	if (topk.size()==k) {
	  break;
	}
      }
    }
  }
  
  delete vectRegIDR;
  delete vectRegIDS;  
  delete heap;  
}

void tests() 
{ 
  //[(0.389,0.075,0.000), (0.389,0.075,0.000)]
  //[(0.622,0.050,0.000), (0.622,0.050,0.000)]
  const double w[3]=
    {
      0.389,0.075,0.000
    };
  const double x[3]=
    {
      0.389,0.075,0.000
    };
  const double y[3]=
    {
      0.622,0.050,0.000
    };
  const double z[3]=
    {
      0.622,0.050,0.000
    };
  
  Point pw=Point(w, 3);
  Point px=Point(x, 3);
  Point py=Point(y, 3);
  Point pz=Point(z, 3);

  PairPoint a=PairPoint(pw, px);
  PairPoint b=PairPoint(py, pz);
  
  cout<<"a "<<a<<endl;
  cout<<"b "<<b<<endl;
  cout<<"dom(a, b) "<<dom(a, b)<<endl;
  cout<<"dom(b, a) "<<dom(b, a)<<endl;

  VectPairPoint vec;
  vec.push_back(a);
  vec.push_back(a);
  vec.push_back(a);
  cout<<"vec "<<vec<<endl;
  
  unsigned dom_flag=SELECT_LEFT|JOIN|SELECT_RIGHT;
  removeDom(vec, b, dom_flag);
  cout<<"vec "<<vec<<endl;  
}

int main(int argc, char *argv[]) {
//   cout<<fixed<<setprecision(3);
//   cerr<<fixed<<setprecision(3);

//   tests();
//   return 0;

  string input_fileR, input_fileS;
  
  if (argc!=4) {
    cerr<<"Usage: "<<argv[0]<<" algorithm input_file query"<<endl;
    return -1;
  }

  RTree::RTree *rtreeR=0;
  IStorageManager *storagemngrR=0;
  StorageManager::IBuffer *fileR=0;

  loadRTree(argv[2], rtreeR, storagemngrR, fileR);

  VectPairPoint vect;
  unsigned k=10;
  query_crt=atoi(argv[3]);

  if (query_crt>=query_no) {
    cerr<<"query should be less than "<<query_no<<endl;
    return -1;
  }

  initRn();

  string alg=string(argv[1]);
  // if (alg=="Original") {
  //   Relax_Skyline(*rtreeR, *rtreeR, vect, SELECT_LEFT|JOIN|SELECT_RIGHT, 0);
  // } 
  // else if (alg=="R-Skyline") {
  //   Relax_Skyline(*rtreeR, *rtreeR, vect, JOIN|SELECT_RIGHT, SELECT_LEFT);
  // } 
  // else if (alg=="S-Skyline") {
  //   Relax_Skyline(*rtreeR, *rtreeR, vect, SELECT_LEFT|JOIN, SELECT_RIGHT);
  // } 
  // else
  if (alg=="J-Skyline") {
    Relax_Skyline(*rtreeR, *rtreeR, vect, SELECT_LEFT|SELECT_RIGHT, JOIN);
  } 
  // else if (alg=="RS-Skyline") {
  //   Relax_Skyline(*rtreeR, *rtreeR, vect, JOIN, SELECT_LEFT|SELECT_RIGHT);
  // } 
  else if (alg=="RJ-Skyline") {
    Relax_Skyline(*rtreeR, *rtreeR, vect, SELECT_RIGHT, SELECT_LEFT|JOIN);
  } 
  else if (alg=="SJ-Skyline") {
    Relax_Skyline(*rtreeR, *rtreeR, vect, SELECT_LEFT, JOIN|SELECT_RIGHT);
  } 
  // else if (alg=="RSJ-Skyline") {
  //   Relax_Skyline(*rtreeR, *rtreeR, vect, 0, SELECT_LEFT|JOIN|SELECT_RIGHT);
  // } 
  // else if (alg=="R-TopK") {
  //   Relax_TopK(*rtreeR, *rtreeR, vect, k, JOIN|SELECT_RIGHT, SELECT_LEFT);
  // } 
  // else if (alg=="S-TopK") {
  //   Relax_TopK(*rtreeR, *rtreeR, vect, k, SELECT_LEFT|JOIN, SELECT_RIGHT);
  // } 
  else if (alg=="J-TopK") {
    Relax_TopK(*rtreeR, *rtreeR, vect, k, SELECT_LEFT|SELECT_RIGHT, JOIN);
  } 
  // else if (alg=="RS-TopK") {
  //   Relax_TopK(*rtreeR, *rtreeR, vect, k, JOIN, SELECT_LEFT|SELECT_RIGHT);
  // } 
  else if (alg=="RJ-TopK") {
    Relax_TopK(*rtreeR, *rtreeR, vect, k, SELECT_RIGHT, SELECT_LEFT|JOIN);
  } 
  else if (alg=="SJ-TopK") {
    Relax_TopK(*rtreeR, *rtreeR, vect, k, SELECT_LEFT, JOIN|SELECT_RIGHT);
  } 
  // else if (alg=="RSJ-TopK") {
  //   Relax_TopK(*rtreeR, *rtreeR, vect, k, 0, SELECT_LEFT|JOIN|SELECT_RIGHT);
  // } 
  else {
    cerr<<"Unknown algorithm "<<argv[1]
	<<". Known algorithms are: "
	<<"Original, "
	// <<"R-Skyline, S-Skyline, J-Skyline, RS-Skyline, RJ-Skyline, SJ-Skyline, RSJ-Skyline, "
	// <<"R-TopK, S-TopK, J-TopK, RS-TopK, RJ-TopK, SJ-TopK, RSJ-TopK"
	<<"J-Skyline, RJ-Skyline, SJ-Skyline, "
	<<"J-TopK, RJ-TopK, SJ-TopK"
	<<"."<<endl;
  }

//   cout<<R[query_crt][1]<<"; "<<R[query_crt][4]<<endl;
//   cout<<Rn[query_crt][1]<<"; "<<Rn[query_crt][4]<<endl;
  cout<<"Number of points: " << vect.size()<<endl;
  cout<<vect<<endl;
  // for (VectPairPointIt it=vect.begin(); it!=vect.end(); it++) {
  //   cout<<*it<<" "<<score(*it)<<endl;
  // }
  // cout<<global_timeIO<<endl;

  if (rtreeR!=0) delete rtreeR;
  if (fileR!=0) delete fileR;
  if (storagemngrR!=0) delete storagemngrR;

  return 0;
}

// ---------------
// ----- OLD -----
// ---------------

/*
 * 1.1 Algorithm 1 - FullIndex_RelaxSelect - V2
 * 
 * - returns skyline
 * 
 * - compute the skyline for each realtion
 * - join the skylines
 *
 */
/*
void FullIndex_RelaxSelect_V2(RTree::RTree& rtreeR, RTree::RTree& rtreeS, VectPairPoint& skyline)
{
  VectPoint *skylineR=new VectPoint();
  VectPoint *skylineS=new VectPoint();
  
  Skyline<greaterRegionID_R_AB, Dom_R_AB>(rtreeR, *skylineR);
  Skyline<greaterRegionID_S_EF, Dom_S_EF>(rtreeS, *skylineS);

//   cout<<skylineR->size()<<endl;
//   cout<<*skylineR<<endl;
//   cout<<skylineS->size()<<endl;
//   cout<<*skylineS<<endl;

  for (VectPointIt itR=skylineR->begin(); itR!=skylineR->end(); itR++) {
    for (VectPointIt itS=skylineS->begin(); itS!=skylineS->end(); itS++) {
      PairPoint pair=PairPoint(*itR, *itS);
      if (cond(pair, JOIN)) {
	skyline.push_back(pair);
      }
    }
  }

  delete skylineR;
  delete skylineS;
}
*/
/*
 * 1.2 Algorithm 2 - FullIndex_RelaxJoin - V2
 *
 * - returns skyline
 *
 * - find all records that satisfy the select condition from the first relation
 * - find all records that satisfy the select condition from the second relation
 *
 * - compute the skyline of the join
 */
/*
void FullIndex_RelaxJoin_V2(RTree::RTree& rtreeR, RTree::RTree& rtreeS, VectPairPoint& skyline) 
{
  VectPoint *selectR=new VectPoint();
  VectPoint *selectS=new VectPoint();
  
  Select<Sel_R_AB>(rtreeR, *selectR);
  Select<Sel_S_EF>(rtreeS, *selectS);  

//   cout<<selectR->size()<<endl;
//   cout<<*selectR<<endl;
//   cout<<selectS->size()<<endl;
//   cout<<*selectS<<endl;

  for (VectPointIt itR=selectR->begin(); itR!=selectR->end(); itR++) {
    for (VectPointIt itS=selectS->begin(); itS!=selectS->end(); itS++) {
      PairPoint pair=PairPoint(*itR, *itS);
      if (!dom(skyline, pair, JOIN)) {
	removeDom(skyline, pair, JOIN);
	skyline.push_back(pair);
      }
    }
  }

  delete selectR;
  delete selectS;
}
*/
/*
 * 1.1 Skyline over one relation
 * 
 * - returns skyline
 *
 * - compute skyline over one relation
 * - compute selection over the other relation
 * - join the two sets of points
 *
 */
/*
void Relax_Skyline_One(RTree::RTree& rtreeR, RTree::RTree& rtreeS, VectPairPoint& skyline, int flag) 
{
  VectPoint *setR=new VectPoint();
  VectPoint *setS=new VectPoint();
  
  if (flag&SELECT_LEFT) {
    Select<Sel_R_01>(rtreeR, *setR);  
  }
  else if (flag&SKYLINE_LEFT) {
    Skyline<greaterRegionID_R_01, Dom_R_01>(rtreeR, *setR);
  }
  
  if (flag&SELECT_RIGHT) {
    Select<Sel_R_46>(rtreeS, *setS);  
  }
  else if (flag&SKYLINE_RIGHT) {
    Skyline<greaterRegionID_R_46, Dom_R_46>(rtreeS, *setS);
  }

//   cout<<setR->size()<<endl;
//   cout<<*setR<<endl;
//   cout<<setS->size()<<endl;
//   cout<<*setS<<endl;

  for (VectPointIt itR=setR->begin(); itR!=setR->end(); itR++) {
    for (VectPointIt itS=setS->begin(); itS!=setS->end(); itS++) {
      PairPoint pair=PairPoint(*itR, *itS);
      if (cond(pair, JOIN)) {
	skyline.push_back(pair);
      }
    }
  }

  delete setR;
  delete setS;  
}
*/
