/*
  $Id: Skyline.cpp 4269 2009-02-09 20:08:05Z rares $

  Copyright (C) 2009 by The Regents of the University of California

  Redistribution of this file is permitted under the terms of the
  BSD license

  Date: 09/01/2005
  Author: Rares Vernica <rares (at) ics.uci.edu> 
*/

#include "Skyline.h"

// ----- Skyline

bool dom(VectPairPoint& vectPairPoint, const PairPoint& pairPoint, const int domFlag)
{
  for (VectPairPointIt it=vectPairPoint.begin(); it!=vectPairPoint.end(); it++) {
    if (dom(*it, pairPoint, domFlag)) {
      return true;
    }
  }
  return false;
}

bool dom(VectPairPoint& vectPairPoint, const PairRegion& pairRegion, const int domFlag)
{
  for (VectPairPointIt it=vectPairPoint.begin(); it!=vectPairPoint.end(); it++) {
    if (dom(*it, pairRegion, domFlag)) {
      return true;
    }
  }
  return false;
}

void removeDom(VectPairPoint& vectPoint, const PairPoint& pairPoint, const int domFlag)
{
  for (VectPairPointIt it=vectPoint.begin(); it!=vectPoint.end();) {
    if (dom(pairPoint, *it, domFlag)) {
      it=vectPoint.erase(it);
    }
    else {
      it++;
    }
  }
}

void removeDom(PairsRegionIDLess& heap, const PairPoint& pairPoint, const int domFlag)
{
  for (PairsRegionIDLessIt it=heap.begin(); it!=heap.end(); it++) {
    if (dom(pairPoint, *it, domFlag)) {
      heap.erase(it);
    }
  }
}

void removeDom(PairsRegionIDLess& heap, const PairRegionID& pairRegionID, const int domFlag)
{
  for (PairsRegionIDLessIt it=heap.begin(); it!=heap.end(); it++) {
    if (dom(pairRegionID, *it, domFlag)) {
      heap.erase(it);
    }
  }
}

// ----- TopK

void prune(PairsRegionIDLessScore* heap, unsigned k) 
{
  if (heap->size()<=k) {
    return;
  }
  PairRegionID pair = *heap->begin();
  unsigned i=0;
  PairsRegionIDLessScoreIt it;
  for (it=heap->begin(); it!=heap->end(); it++, i++) {
    if (i<k) {
      continue;
    }
    else if (i==k) {
      pair=*it;
    }
    else {
      if (scoreMinMax(pair)>scoreMin(*it)) {
	break;
      }
    }
  }
  if (it!=heap->end()) {
    heap->erase(it, heap->end());
  }
}
