/*
  $Id: Skyline.h 4269 2009-02-09 20:08:05Z rares $

  Copyright (C) 2009 by The Regents of the University of California

  Redistribution of this file is permitted under the terms of the
  BSD license

  Date: 09/01/2005
  Author: Rares Vernica <rares (at) ics.uci.edu> 
*/

#ifndef __Skyline_h
#define __Skyline_h

#include <iostream>
#include <iterator>
#include <vector>

using namespace std;

#include "DataStruct.h"
#include "DataSetUtil.h"
#include "DataCont.h"
#include "RTreeUtil.h"

/*
 * - skyline
 *   - list of pairs
 *   - verify if a pair is dominated by any pair in the skyline
 */

/* ----- skyline ----- */

bool dom(VectPairPoint& vectPairPoint, 
	 const PairPoint& pairPoint, 
	 const int domFlag=SELECT_LEFT|JOIN|SELECT_RIGHT);

bool dom(VectPairPoint& vectPairPoint, 
	 const PairRegion& pairRegion, 
	 const int domFlag=SELECT_LEFT|JOIN|SELECT_RIGHT);

inline bool dom(VectPairPoint& vectPairPoint,
		const PairRegionID& pairRegionID,
		const int domFlag=SELECT_LEFT|JOIN|SELECT_RIGHT) 
{
  return dom(vectPairPoint, PairRegion(pairRegionID.first.getRegion(), pairRegionID.second.getRegion()), domFlag);
}

template <class domPointRegion> bool dom(VectPoint& vectPoint, const Region& region)
{
  for (VectPointIt it=vectPoint.begin(); it!=vectPoint.end(); it++) {
    if (domPointRegion()(*it, region)) {
      return true;
    }
  }
  return false;
}

void removeDom(VectPairPoint& vectPoint, const PairPoint& pairPoint, const int domFlag);
void removeDom(PairsRegionIDLess& heap, const PairPoint& pairPoint, const int domFlag);
void removeDom(PairsRegionIDLess& heap, const PairRegionID& pairRegionID, const int domFlag);

/*
template <class domPointPoint> void removeDom(VectPoint& vectPoint, const Point& point)
{
  for (VectPointIt it=vectPoint.begin(); it!=vectPoint.end();) {
    if (domPointPoint()(point, *it)) {
      it=vectPoint.erase(it);
    }
    else {
      it++;
    }
  }
}
*/

// ----- Select

template <class Sel> void Select(RTree::RTree& rtree, VectPoint& select)
{
  typedef queue<RegionID> HeapRegionID;
  HeapRegionID *heap=new HeapRegionID();
  RegionID regID;

  regID=RegionID(rtree.m_rootID, true, rtree.readNode(rtree.m_rootID)->m_nodeMBR);
  global_timeIO++;
  if (Sel()(regID.getRegion())) {
    heap->push(regID);
  }
  while (!heap->empty()) {
    regID=heap->front();
    heap->pop();
    if (regID.index) {
      RTree::NodePtr node=rtree.readNode(regID.id);
      global_timeIO++;
      for (unsigned long cChild=0; cChild<node->getChildrenCount(); cChild++) {
	regID=RegionID(node->getChildIdentifier(cChild), node->isIndex(), *node->m_ptrMBR[cChild]);
	if (Sel()(regID.getRegion())) {
	  heap->push(regID);
	}
      }
    }
    else {
      Point point=Point(regID.getRegion().m_pLow, regID.getRegion().m_dimension);
      select.push_back(point);
    }
  }

  delete heap;
}

// ----- Skyline
/*
template <class greaterRegion, class Dom> void Skyline(RTree::RTree& rtree, VectPoint& skyline) 
{
  typedef priority_queue<RegionID, vector<RegionID>, greaterRegion> HeapRegionID;
  HeapRegionID *heap=new HeapRegionID();
  RegionID regID;

  regID=RegionID(rtree.m_rootID, true, rtree.readNode(rtree.m_rootID)->m_nodeMBR);
  global_timeIO++;
  heap->push(regID);
  while (!heap->empty()) {
    regID=heap->top();
    heap->pop();
    if (!dom<Dom>(skyline, regID.getRegion())) {
      if (regID.index) {
	RTree::NodePtr node=rtree.readNode(regID.id);
	global_timeIO++;
	for (unsigned long cChild=0; cChild<node->getChildrenCount(); cChild++) {
	  regID=RegionID(node->getChildIdentifier(cChild), node->isIndex(), *node->m_ptrMBR[cChild]);
	  if (!dom<Dom>(skyline, regID.getRegion())) {
	    heap->push(regID);
	  }
	}
      }
      else {
	Point point=Point(regID.getRegion().m_pLow, regID.getRegion().m_dimension);
	removeDom<Dom>(skyline, point); // remove form the skyline pairs that are dominated by the new pair
	skyline.push_back(point);
      }
    }
  }

  delete heap;
}
*/
// ----- TopK

void prune(PairsRegionIDLessScore *heap, unsigned k);

#endif
