/*
  $Id: Dataset.h 4269 2009-02-09 20:08:05Z rares $

  Copyright (C) 2009 by The Regents of the University of California

  Redistribution of this file is permitted under the terms of the
  BSD license

  Date: 09/01/2005
  Author: Rares Vernica <rares (at) ics.uci.edu> 
*/

#ifndef _Dataset_
#define _Dataset_

//--

#include <cmath>
#include <iostream>
#include <iterator>
#include <set>
#include <vector>
#include <tr1/unordered_map>

using namespace std;
using namespace tr1;

//--

template <class Tuple>
unsigned getTuple_attrJ(const Tuple& tuple) 
{
  return tuple.attrJ;
}

template <class Tuple>  
class SortPredA
{
private:
  const vector<Tuple> *dataset;
public:
  SortPredA(const vector<Tuple> &ds): dataset(&ds) { }
  bool operator()(const unsigned idX, const unsigned idY) const
  {
    return (*dataset)[idX].relaxA() < (*dataset)[idY].relaxA();
  }
};

template <class Tuple>  
class SortPredB
{
private:
  const vector<Tuple> *dataset;
public:
  SortPredB(const vector<Tuple> &ds): dataset(&ds) { }
  bool operator()(const unsigned idX, const unsigned idY) const
  {
    return (*dataset)[idX].relaxB() < (*dataset)[idY].relaxB();
  }
};

template <class Tuple>  
class SortPredC
{
private:
  const vector<Tuple> *dataset;
public:
  SortPredC(const vector<Tuple> &ds): dataset(&ds) { }
  bool operator()(const unsigned idX, const unsigned idY) const
  {
    return (*dataset)[idX].relaxC() < (*dataset)[idY].relaxC();
  }
};

//-- IO

template <class Tuple>
ostream& operator<<(ostream&  out, const vector<Tuple> &vect) 
{
  copy(vect.begin(), vect.end(), ostream_iterator<Tuple>(out, "\n"));
  return out; 
}		 

//--

typedef unordered_multimap<unsigned, unsigned> HashIdxUnsigned;
typedef HashIdxUnsigned::const_iterator HashIdxUnsignedCIt;
typedef pair<HashIdxUnsignedCIt, HashIdxUnsignedCIt> PairHashIdxUnsignedCIt;

typedef vector<unsigned> VectUnsigned;
typedef pair<unsigned, unsigned> PairUnsigned;
typedef vector<PairUnsigned> VectPairUnsigned;
typedef set<PairUnsigned> SetPairUnsigned;
typedef VectUnsigned::const_iterator VectUnsignedCIt;
typedef VectPairUnsigned::const_iterator VectPairUnsignedCIt;

//-- IO

inline ostream& operator<<(ostream&  out, const VectUnsigned &vect) 
{
  copy(vect.begin(), vect.end(), ostream_iterator<unsigned>(out, "\n"));
  return out; 
}		 

inline ostream& operator<<(ostream &out, const PairUnsigned &pair)
{
  return out << pair.first << "\t" << pair.second;
}

inline ostream& operator<<(ostream &out, const VectPairUnsigned &vect)
{
  for (VectPairUnsigned::const_iterator it = vect.begin(); it != vect.end(); ++it)
    {
      out << *it << endl;
    }
  return out;
}

//--

class Inc
{
private:
  unsigned i;
public:
  Inc(): i(0) { }
  void reset()
  {
    i = 0;
  }
  unsigned operator()()
  {
    return i++;
  }
};

//--

#endif
