/*
  $Id: DatasetIMDB.h 4269 2009-02-09 20:08:05Z rares $

  Copyright (C) 2009 by The Regents of the University of California

  Redistribution of this file is permitted under the terms of the
  BSD license

  Date: 09/01/2005
  Author: Rares Vernica <rares (at) ics.uci.edu> 
*/

#ifndef _DatasetIMDB_
#define _DatasetIMDB_

#import <string>   

using namespace std;

// const string datasets_path         = "/extra/rvernica0/projects/datasets/";
// const string datasets_directory[1] = { "imdb/" };

const string datasets_path         = "";
const string datasets_directory[1] = { "" };

const string datasets_fileext      = ".txt";

#ifdef DIM_1_X_ONLY
const string datasetR_filename    = "movies.id_year";
#endif

#ifdef DIM_X_1_ONLY
const string datasetS_filename    = "actors.movie_id_datebirth_year";
#endif

#ifdef DIM_2_X_ONLY
// const string datasetR_filename    = "movies.id_year_rank";
const string datasetR_filename    = "movies.movid_year_rank";
#endif

#ifdef DIM_X_2_ONLY
// const string datasetS_filename    = "actors.movie_id_datebirth_year_billingpos";
const string datasetS_filename    = "actors.movid_dobyear_billp";
#endif

#ifdef DIM_3_X_ONLY
const string datasetR_filename    = "movies.id_year_rank_time";
#endif

#ifdef DIM_X_3_ONLY
const string datasetS_filename    = "actors.movie_id_datebirth_year_billingpos_datebirth_month";
#endif

const unsigned no_runs = 1;

//-- Movies & Actors

/*x
  SELECT * FROM imdb_movies, imdb_actors
  WHERE imdb_movies.year = ???? AND
        imdb_movies.rank = ?.? AND
        imdb_movies.id = imdb_actors.movie_id AND
	imdb_actors.datebirth_year = ???? AND
	imdb_actors.billingpos = ?;
*/

class TupleMovie // imdb_movies
{
public:
  int attrA; // year
  unsigned attrJ; // id - join attribute

  static int qattrA;
  int relaxA() const { return abs(attrA - qattrA); }
  static const int tupleR_qattrA[];

#ifdef DIM_2_X
  float attrB; // rank
  static float qattrB;
  float relaxB() const { return fabs(attrB - qattrB); }
  static const float tupleR_qattrB[];
#endif

#ifdef DIM_3_X
  int attrC; // time
  static int qattrC;
  float relaxC() const { return abs(attrC - qattrC); }
  static const int tupleR_qattrC[];
#endif

#ifdef TopK
  float score() const
  {
    return 
      weights[0] * relaxA()
#ifdef DIM_2_X
      + weights[1] * relaxB()
#endif
#ifdef DIM_3_X
      + weights[2] * relaxC()
#endif
      ;
  }
#endif
};

int TupleMovie::qattrA;
const int TupleMovie::tupleR_qattrA[] = { 1984 /*, 1988, 1992, 1994 */};

#ifdef DIM_2_X
float TupleMovie::qattrB;
const float TupleMovie::tupleR_qattrB[] = { 7 /* 3, 7 */};
#endif

#ifdef DIM_3_X
int TupleMovie::qattrC;
const int TupleMovie::tupleR_qattrC[] = { 120 };
#endif

class TupleActor // imdb_actors
{
public:
  int attrA; // datebirth_year
  unsigned attrJ; // movie_id - join attribute

  static int qattrA;
  int relaxA() const { return abs(attrA - qattrA); }
  static const int tupleS_qattrA[];

#ifdef DIM_X_2
  int attrB; // billingpos
  static int qattrB;
  int relaxB() const { return abs(attrB - qattrB); }
  static const int tupleS_qattrB[];
#endif

#ifdef DIM_X_3
  int attrC; // datebirth_month
  static int qattrC;
  int relaxC() const { return abs(attrC - qattrC); }
  static const int tupleS_qattrC[];
#endif

#ifdef TopK
  float score() const
  {
    return 
      weights[DIM_S_A] * relaxA()
#ifdef DIM_X_2
      + weights[DIM_S_A + 1] * relaxB()
#endif
#ifdef DIM_X_3
      + weights[DIM_S_A + 2] * relaxC()
#endif
      ;
  }
#endif
};

int TupleActor::qattrA;
const int TupleActor::tupleS_qattrA[] = { 1950 /*, 1954, 1958, 1962 */};

#ifdef DIM_X_2
int TupleActor::qattrB;
const int TupleActor::tupleS_qattrB[] = { 2 /*, 6 */};
#endif

#ifdef DIM_X_3
int TupleActor::qattrC;
const int TupleActor::tupleS_qattrC[] = { 4 };
#endif

//-- IO

inline istream& operator>>(istream &in, TupleMovie &tuple)
{
  in >> tuple.attrJ >> tuple.attrA;

#ifdef DIM_2_X
  in >> tuple.attrB;
#endif

#ifdef DIM_3_X
  in >> tuple.attrC;
#endif

  return in;
  
}

inline ostream& operator<<(ostream &out, const TupleMovie &tuple)
{
  out << tuple.attrJ << "\t" << tuple.attrA;

#ifdef DIM_2_X
  out << "\t" <<  tuple.attrB;
#endif

#ifdef DIM_3_X
  out << "\t" <<  tuple.attrC;
#endif

#ifdef TopK
  out << "\t" << tuple.score();
#endif

  return out;
}

inline istream& operator>>(istream &in, TupleActor &tuple)
{
  in >> tuple.attrJ >> tuple.attrA;

#ifdef DIM_X_2
  in >> tuple.attrB;
#endif

#ifdef DIM_X_3
  in >> tuple.attrC;
#endif

  return in;
}  

inline ostream& operator<<(ostream &out, const TupleActor &tuple)
{
  out << tuple.attrJ << "\t" << tuple.attrA;

#ifdef DIM_X_2
  out << "\t" <<  tuple.attrB;
#endif

#ifdef DIM_X_3
  out << "\t" <<  tuple.attrC;
#endif

#ifdef TopK
  out << "\t" << tuple.score();
#endif

  return out;
}

//--

typedef vector<TupleMovie> VectTupleR;
typedef vector<TupleActor> VectTupleS;

#endif
