/*
  $Id: DatasetSynth.h 4269 2009-02-09 20:08:05Z rares $

  Copyright (C) 2009 by The Regents of the University of California

  Redistribution of this file is permitted under the terms of the
  BSD license

  Date: 09/01/2005
  Author: Rares Vernica <rares (at) ics.uci.edu> 
*/

#ifndef _DatasetSynth_
#define _DatasetSynth_

#import <string>   

using namespace std;

const string datasets_path         = "/extra/rvernica0/projects/datasets/synth/";
const string datasets_directory[3] = { "corr/", "anti-corr/", "indep/" };
const string datasets_fileext      = ".txt";

const string datasetR_filename    = "datasetR";
const string datasetS_filename    = "datasetS";

const unsigned no_runs = 10;

//-- Synth - R & S

/*
  SELECT * FROM R, S
  WHERE R.A = 0 AND
        R.B = 0 AND
        R.J = S.J AND
	S.A = 0 AND
	S.B = 0;
*/

class TupleR // synth_movies
{
public:
  float attrA; // A
  unsigned attrJ; // J - join attribute

  static float qattrA;
  float relaxA() const { return fabs(attrA - qattrA); }
  const static float tupleR_qattrA[];

#ifdef DIM_2_X
  float attrB; // B
  static float qattrB;
  float relaxB() const { return fabs(attrB - qattrB); }
  const static float tupleR_qattrB[];
#endif

#ifdef DIM_3_X
  float attrC; // C
  static float qattrC;
  float relaxC() const { return fabs(attrC - qattrC); }
  const static float tupleR_qattrC[];
#endif

#ifdef TopK
  float score() const
  {
    return 
      weights[0] * relaxA()
#ifdef DIM_2_X
      + weights[1] * relaxB()
#endif
#ifdef DIM_3_X
      + weights[2] * relaxC()
#endif
      ;
  }
#endif
};

float TupleR::qattrA;
const float TupleR::tupleR_qattrA[] = { 0 };

#ifdef DIM_2_X
float TupleR::qattrB;
const float TupleR::tupleR_qattrB[] = { 0 };
#endif

#ifdef DIM_3_X
float TupleR::qattrC;
const float TupleR::tupleR_qattrC[] = { 0 };
#endif

class TupleS // imdb_actors
{
public:
  float attrA; // A
  unsigned attrJ; // J - join attribute

  static float qattrA;
  float relaxA() const { return fabs(attrA - qattrA); }
  const static float tupleS_qattrA[];

#ifdef DIM_X_2
  float attrB; // B
  static float qattrB;
  float relaxB() const { return fabs(attrB - qattrB); }
  const static float tupleS_qattrB[];
#endif

#ifdef DIM_X_3
  float attrC; // C
  static float qattrC;
  float relaxC() const { return fabs(attrC - qattrC); }
  const static float tupleS_qattrC[];
#endif

#ifdef TopK
  float score() const
  {
    return 
      weights[DIM_S_A] * relaxA()
#ifdef DIM_X_2
      + weights[DIM_S_A + 1] * relaxB()
#endif
#ifdef DIM_X_3
      + weights[DIM_S_A + 2] * relaxC()
#endif
      ;
  }
#endif
};

float TupleS::qattrA;
const float TupleS::tupleS_qattrA[] = { 0 };

#ifdef DIM_X_2
float TupleS::qattrB;
const float TupleS::tupleS_qattrB[] = { 0 };
#endif

#ifdef DIM_X_3
float TupleS::qattrC;
const float TupleS::tupleS_qattrC[] = { 0 };
#endif

//-- IO

inline istream& operator>>(istream &in, TupleR &tuple)
{
  in >> tuple.attrJ >> tuple.attrA;

#ifdef DIM_2_X
  in >> tuple.attrB;
#endif

#ifdef DIM_3_X
  in >> tuple.attrC;
#endif

  return in;
  
}

inline ostream& operator<<(ostream &out, const TupleR &tuple)
{
  out << tuple.attrJ << "\t" << tuple.attrA;

#ifdef DIM_2_X
  out << "\t" <<  tuple.attrB;
#endif

#ifdef DIM_3_X
  out << "\t" <<  tuple.attrC;
#endif

#ifdef TopK
  out << "\t" << tuple.score();
#endif

  return out;
}

inline istream& operator>>(istream &in, TupleS &tuple)
{
  in >> tuple.attrJ >> tuple.attrA;

#ifdef DIM_X_2
  in >> tuple.attrB;
#endif

#ifdef DIM_X_3
  in >> tuple.attrC;
#endif

  return in;
}  

inline ostream& operator<<(ostream &out, const TupleS &tuple)
{
  out << tuple.attrJ << "\t" << tuple.attrA;

#ifdef DIM_X_2
  out << "\t" <<  tuple.attrB;
#endif

#ifdef DIM_X_3
  out << "\t" <<  tuple.attrC;
#endif

#ifdef TopK
  out << "\t" << tuple.score();
#endif

  return out;
}

//--

typedef vector<TupleR> VectTupleR;
typedef vector<TupleS> VectTupleS;

#endif
